local SelectCriterion, parent = torch.class('nnf.SelectCriterion', 'nn.Criterion')

function SelectCriterion:__init(crit)
  parent.__init(self)
  self.gradInput = {}
  self.criterion = crit
end

-- assumes target is {select_classes, other_targets_for_selected_class}
-- input is a table
function SelectCriterion:updateOutput(input, target)
  local indexes = target[1]
  local otarget = target[2]
  self.output = 0
  local num = 0
  for i = 1, indexes:size(1) do
    local idx = indexes[i] - 1
    if idx > 0 then
      self.output = self.output + self.criterion:updateOutput(input[idx][i], otarget[i])
      num = num + 1
    end
  end
  self.output = self.output/num
  return self.output
end

function SelectCriterion:updateGradInput(input, target)
  local indexes = target[1]
  local otarget = target[2]
  for i = 1, #input do
    self.gradInput[i] = self.gradInput[i] or input[i].new()
    self.gradInput[i]:resizeAs(input[i])
    self.gradInput[i]:zero()
  end
  local num = 0
  for i = 1, indexes:size(1) do
    local idx = indexes[i] - 1
    if idx > 0 then
      self.gradInput[idx][i]:add(self.criterion:updateGradInput(input[idx][i], otarget[i]))
      num = num + 1
    end
  end
  for i = 1, #input do
    self.gradInput[i]:div(num)
  end
  return self.gradInput
end
