require 'image'
require 'paths'
--require 'trepl'

local M = {}
local MixBP = torch.class('MixBP', M)

function MixBP:__init(bp, imlist, targets, rootfolder, num_imgs)
  self.bp = bp
  self.image_transformer = bp.feat_provider.image_transformer
  self.imlist = imlist
  self.targets = targets
  self.rootfolder = rootfolder
  self.num_imgs = num_imgs
end

function MixBP:getBatch()
  local batches, targets = self.bp:getBatch()
  self._batches = self._batches or {torch.FloatTensor(), torch.FloatTensor()}
  self._targets = self._targets or torch.FloatTensor()
  self._batches[1]:resize(batches[1]:size(1)+1, table.unpack(batches[1][1]:size():totable()))
  self._batches[2]:resize(batches[2]:size(1)+self.num_imgs, batches[2]:size(2))
  self._targets:resize(targets:size(1)+self.num_imgs)
  
  self._batches[1]:narrow(1,1,batches[1]:size(1)):copy(batches[1])
  self._batches[2]:narrow(1,1,batches[2]:size(1)):copy(batches[2])
  self._targets:narrow(1,1,targets:size(1)):copy(targets)
  
  self:_getBatch(batches[1]:size(), batches[2]:size())
  --local nbatches, ntargets = self:_getBatch(batches[1]:size(), batches[2]:size())
  --nbatches[1] = torch.FloatTensor.cat(batches[1], nbatches[1], 1)
  --nbatches[2] = torch.FloatTensor.cat(batches[2], nbatches[2], 1)
  --ntargets = torch.FloatTensor.cat(targets, ntargets, 1)
  --return nbatches, ntargets
  self._counter = self._counter or 0
  self._counter = self._counter + 1
  if self._counter % 50 == 0 then
    collectgarbage()
  end
  return self._batches, self._targets
end

function MixBP:_getBatch(bsize, lsize)
  local imlist = self.imlist
  local num_imgs = self.num_imgs
  local rootfolder = self.rootfolder
  local targets = self.targets
  local im_size_x = bsize[4]
  local im_size_y = bsize[3]
  local imgs_per_batch = bsize[1]
  local image_transformer = self.image_transformer

  local context = 1.15
  local grid_size = math.sqrt(num_imgs)
  local step_size_x = im_size_x/grid_size
  local step_size_y = im_size_y/grid_size
  
  local min_step_size = math.min(step_size_x,step_size_y)
  
  local selected_imgs = torch.LongTensor(num_imgs):random(1, #imlist)
  local montage = self._batches[1]:narrow(1,imgs_per_batch + 1,1):zero()--torch.FloatTensor(1,3,im_size_y,im_size_x):zero()
  local boxes = self._batches[2]:narrow(1,lsize[1] + 1, num_imgs)--torch.FloatTensor(num_imgs, 5)
  local labels = self._targets:narrow(1,lsize[1] + 1, num_imgs) --torch.FloatTensor(num_imgs)
  for i = 1, num_imgs do
    local select_img = selected_imgs[i]
    local im = image.load(paths.concat(rootfolder, imlist[select_img]), 3, 'float')
    im = image_transformer:preprocess(im)
    --print(im:size(),imlist[select_img])
    im = image.scale(im, min_step_size/context)
    local ims = im[1]:size()
    local x = (i-1)%grid_size
    local y = math.floor((i-1)/grid_size)
    local offset_x = (step_size_x-ims[2])/2
    local offset_y = (step_size_y-ims[1])/2
    local lmont = montage:narrow(3,y*step_size_y+1 + offset_y, ims[1]):narrow(4,x*step_size_x+1 + offset_x, ims[2])
    lmont:copy(im)
    local bb = boxes[i]
    bb[1] = imgs_per_batch + 1
    bb[2] = x*step_size_x+1 + offset_x
    bb[3] = y*step_size_y+1 + offset_y
    bb[4] = x*step_size_x+1 + ims[2] + offset_x
    bb[5] = y*step_size_y+1 + ims[1] + offset_y
    labels[i] = targets[select_img]-- + 1 -- bg
  end
  --return {montage, boxes} , labels
end

function MixBP:__tostring()
  local str = ('%s\n  original: %s\n  path: %s\n  montage size: %d'):format(torch.typename(self), tostring(self.bp), self.rootfolder, self.num_imgs)
  return str
end

return M.MixBP

--[[
-- slightly modified to also read the targets
function lines_from(file)
  -- get all lines from a file, returns an empty 
  -- list/table if the file does not exist
  if not paths.filep(file) then return {} end
  local lines = {}
  local targets = {}
  for line in io.lines(file) do
    local l = {}
    for i in string.gmatch(line, "%S+") do
      table.insert(l, i)
    end
    table.insert(lines,l[1])
    table.insert(targets,tonumber(l[2]))
  end
  return lines, targets
end


basefolder = '/media/francisco/Seagate Expansion Drive/francisco/work/datasets/render4cnn'

imlist, targets = lines_from(paths.concat(basefolder,'imglist.txt'))
--]]
--Ims, boxes, labels = make_montage(imlist, 9, basefolder, targets)
--image.display(Ims)

