local utils = paths.dofile('../utils.lua')
local nms = paths.dofile('../nms.lua')

local keep_top_k = utils.keep_top_k
local boxoverlap = utils.boxoverlap
local VOCap = utils.VOCap


local pose_utils = {}

local function flip_angle(x)
  return (-x)%360
end

function pose_utils.VOCevaldet_pose(dataset,scored_boxes,cls, nPoses, discretize)
  local num_pr = 0
  local energy = {}
  local correct = {}
  local angle_correct = {}
  
  local divAng = 360/nPoses
  local count = 0
  
  for i=1,dataset:size() do   
    local ann = dataset:getAnnotation(i)   
    local bbox = {}
    local det = {}
    local angle = {}
    for idx,obj in ipairs(ann.object) do
      if obj.name == cls and obj.difficult == '0' then
        table.insert(bbox,{obj.bndbox.xmin,obj.bndbox.ymin,
                           obj.bndbox.xmax,obj.bndbox.ymax})
        table.insert(det,0)
        local ang
        if obj.viewpoint.distance == '0' then
          ang = obj.viewpoint.azimuth_coarse
        else
          ang = obj.viewpoint.azimuth
        end
        if false then
          ang = flip_angle(ang)
          --[[
          local box = bbox[#bbox]
          local tt = box[1]
          local im_width = ann.size.width
          box[1] = im_width-box[3]+1
          box[3] = im_width-tt    +1
          --]]
        end
        -- discretize it
        ang = math.fmod(math.ceil((ang-divAng/2)/divAng),nPoses)+1
        table.insert(angle, ang)
        count = count + 1
      end
    end
    
    bbox = torch.Tensor(bbox)
    det = torch.Tensor(det)
    angle = torch.Tensor(angle)
    
    local num = scored_boxes[i]:numel()>0 and scored_boxes[i]:size(1) or 0
    for j=1,num do
      local bbox_pred = scored_boxes[i][j]
      num_pr = num_pr + 1
      table.insert(energy,bbox_pred[6])
      
      if bbox:numel()>0 then
        local o = boxoverlap(bbox,bbox_pred[{{1,4}}])
        local maxo,index = o:max(1)
        maxo = maxo[1]
        index = index[1]
        if maxo >=0.5 and det[index] == 0 then
          correct[num_pr] = 1
          det[index] = 1
          local pred_angle = bbox_pred[5]
          if discretize then
            pred_angle = math.fmod(math.ceil((pred_angle-divAng/2)/divAng),nPoses)+1
          end
          if angle[index] == pred_angle then
            angle_correct[num_pr] = 1
          else
            angle_correct[num_pr] = 0
          end
        else
          correct[num_pr] = 0
          angle_correct[num_pr] = 0
        end
      else
          correct[num_pr] = 0        
          angle_correct[num_pr] = 0
      end
    end
    
  end
  
  if #energy == 0 then
    return 0,torch.Tensor(),torch.Tensor(), torch.Tensor()
  end
  
  energy = torch.Tensor(energy)
  correct = torch.Tensor(correct)
  angle_correct = torch.Tensor(angle_correct)
  
  local threshold,index = energy:sort(true)

  correct = correct:index(1,index)
  angle_correct = angle_correct:index(1,index)

  local n = threshold:numel()
  
  local recall = torch.zeros(n)
  local precision = torch.zeros(n)
  local accuracy = torch.zeros(n)

  local num_correct = 0
  local num_correct_view = 0

  for i = 1,n do
      --compute precision
      num_positive = i
      num_correct = num_correct + correct[i]
      if num_positive ~= 0 then
          precision[i] = num_correct / num_positive;
      else
          precision[i] = 0;
      end
      
      num_correct_view = num_correct_view + angle_correct[i]
      if num_positive ~= 0 then
          accuracy[i] = num_correct_view / num_positive;
      else
          accuracy[i] = 0;
      end

      --compute recall
      recall[i] = num_correct / count
  end

  local ap, avp
  if dataset.dataset_name == 'VOC2007' then
    ap = VOC2007ap(recall, precision)
    avp = VOC2007ap(recall, accuracy)
  else
    ap = VOCap(recall, precision)
    avp = VOCap(recall, accuracy)
  end

  io.write(('AP = %.4f\n'):format(ap));

  return ap, avp, recall, precision, accuracy
end

function pose_utils.pruneDetections(self, aboxes, thresh, output, boxes, i, max_per_image, max_per_set)
  local nPoses = self.nPoses
  local add_bg = 1
  self._scored_boxes = self._scored_boxes or torch.FloatTensor()
  local scored_boxes = self._scored_boxes
  local dataset = self.dataset
  -- do a NMS for each class, based on the scores from the classifier
  for j=1,dataset.num_classes do
    local scores = output:narrow(2,(j-1)*nPoses+add_bg + 1, nPoses):sum(2):select(2,1)
    local _,best_angles = output:narrow(2,(j-1)*nPoses+add_bg + 1, nPoses):max(2)
    -- only select detections with a score greater than thresh
    -- this avoid doing NMS on too many bboxes with low score
    local idx = torch.range(1,scores:numel()):long()
    local idx2 = scores:gt(thresh[j])
    idx = idx[idx2]
    scored_boxes:resize(idx:numel(),6)
    if scored_boxes:numel() > 0 then
      scored_boxes:narrow(2,1,4):index(boxes,1,idx)
      scored_boxes:select(2,6):copy(scores[idx2])
      scored_boxes:select(2,5):copy(best_angles[idx2])
    end
    local keep = nms(scored_boxes,0.3)
    if keep:numel()>0 then
      local _,ord = torch.sort(scored_boxes:select(2,6):index(1,keep),true)
      ord = ord:narrow(1,1,math.min(ord:numel(),max_per_image))
      keep = keep:index(1,ord)
      aboxes[j][i] = scored_boxes:index(1,keep)
    else
      aboxes[j][i] = torch.FloatTensor()
    end

    -- remove low scoring boxes and update threshold
    if i%1000 == 0 then
      aboxes[j],thresh[j] = keep_top_k(aboxes[j],max_per_set)
    end

  end
end

function pose_utils.pruneDetectionsMixed(self, aboxes, thresh, output, boxes, i, max_per_image, max_per_set)
  local nPoses = self.nPoses
  local manif_size = self.manif_size
  --local a_i = torch.linspace(0,90,manif_size):mul(math.pi/180):data()
  local a_i = torch.linspace(-60,60,manif_size):mul(math.pi/180):data()
  local add_bg = 1
  self._scored_boxes = self._scored_boxes or torch.FloatTensor()
  local scored_boxes = self._scored_boxes
  local dataset = self.dataset
  local oscore = output[1]
  local ang_pred = output[2]
  local best_angles = torch.FloatTensor(oscore:size(1))
  local best_angles_c = best_angles:data()
  -- do a NMS for each class, based on the scores from the classifier
  for j=1,dataset.num_classes do
    local scores = oscore:select(2,j+add_bg)
    local angs = ang_pred[j]--output:narrow(2,(j-1)*nPoses+add_bg + 1, nPoses):max(2)
    local angs_c = angs:data()
    if manif_size == 2 then
      for cc = 0, angs:size(1)-1 do
        best_angles_c[cc] = (math.atan2(angs_c[2*cc+1],angs_c[2*cc])*180/math.pi)%360
      end
    else
      for cc = 0, angs:size(1)-1 do
        local y1 = 0
        local x1 = 0
        for dim = 0, manif_size-1 do
          y1 = y1 + angs_c[manif_size*cc+dim]*math.cos(a_i[dim])
          x1 = x1 + angs_c[manif_size*cc+dim]*math.sin(a_i[dim])
        end
        -- check the order
        best_angles_c[cc] = (math.atan2(x1,y1)*180/math.pi)%360
        --best_angles_c[cc] = (math.atan2(angs_c[2*cc+1],angs_c[2*cc])*180/math.pi)%360
      end
    end
    -- only select detections with a score greater than thresh
    -- this avoid doing NMS on too many bboxes with low score
    local idx = torch.range(1,scores:numel()):long()
    local idx2 = scores:gt(thresh[j])
    idx = idx[idx2]
    scored_boxes:resize(idx:numel(),6)
    if scored_boxes:numel() > 0 then
      scored_boxes:narrow(2,1,4):index(boxes,1,idx)
      scored_boxes:select(2,6):copy(scores[idx2])
      scored_boxes:select(2,5):copy(best_angles[idx2])
    end
    local keep = nms(scored_boxes,0.3)
    if keep:numel()>0 then
      local _,ord = torch.sort(scored_boxes:select(2,6):index(1,keep),true)
      ord = ord:narrow(1,1,math.min(ord:numel(),max_per_image))
      keep = keep:index(1,ord)
      aboxes[j][i] = scored_boxes:index(1,keep)
    else
      aboxes[j][i] = torch.FloatTensor()
    end

    -- remove low scoring boxes and update threshold
    if i%1000 == 0 then
      aboxes[j],thresh[j] = keep_top_k(aboxes[j],max_per_set)
    end

  end
end

function pose_utils.pruneDetectionsMixedSepClassif(self, aboxes, thresh, output, boxes, i, max_per_image, max_per_set)
  local nPoses = self.nPoses
  local manif_size = self.manif_size
  --local a_i = torch.linspace(0,90,manif_size):mul(math.pi/180):data()
  local a_i = torch.linspace(-60,60,manif_size):mul(math.pi/180):data()
  local add_bg = 1
  local frcnn_detec = self.frcnn_detec
  local add_bg = self.add_bg
  local dataset = self.dataset
  local I = dataset:getImage(i)
  local classif_scores = frcnn_detec:detect(I, boxes)
  classif_scores = classif_scores:index(2,torch.LongTensor{1,2,3,5,6,7,8,10,12,15,19,20,21})

  self._scored_boxes = self._scored_boxes or torch.FloatTensor()
  local scored_boxes = self._scored_boxes
  local dataset = self.dataset
  local oscore = output[1]
  local ang_pred = output[2]
  local best_angles = torch.FloatTensor(oscore:size(1))
  local best_angles_c = best_angles:data()
  -- do a NMS for each class, based on the scores from the classifier
  for j=1,dataset.num_classes do
    local scores = classif_scores:select(2,j+1) --oscore:select(2,j+add_bg)
    local angs = ang_pred[j]--output:narrow(2,(j-1)*nPoses+add_bg + 1, nPoses):max(2)
    local angs_c = angs:data()
    if manif_size == 2 then
      for cc = 0, angs:size(1)-1 do
        best_angles_c[cc] = (math.atan2(angs_c[2*cc+1],angs_c[2*cc])*180/math.pi)%360
      end
    else
      for cc = 0, angs:size(1)-1 do
        local y1 = 0
        local x1 = 0
        for dim = 0, manif_size-1 do
          y1 = y1 + angs_c[manif_size*cc+dim]*math.cos(a_i[dim])
          x1 = x1 + angs_c[manif_size*cc+dim]*math.sin(a_i[dim])
        end
        -- check the order
        best_angles_c[cc] = (math.atan2(x1,y1)*180/math.pi)%360
        --best_angles_c[cc] = (math.atan2(angs_c[2*cc+1],angs_c[2*cc])*180/math.pi)%360
      end
    end
    -- only select detections with a score greater than thresh
    -- this avoid doing NMS on too many bboxes with low score
    local idx = torch.range(1,scores:numel()):long()
    local idx2 = scores:gt(thresh[j])
    idx = idx[idx2]
    scored_boxes:resize(idx:numel(),6)
    if scored_boxes:numel() > 0 then
      scored_boxes:narrow(2,1,4):index(boxes,1,idx)
      scored_boxes:select(2,6):copy(scores[idx2])
      scored_boxes:select(2,5):copy(best_angles[idx2])
    end
    local keep = nms(scored_boxes,0.3)
    if keep:numel()>0 then
      local _,ord = torch.sort(scored_boxes:select(2,6):index(1,keep),true)
      ord = ord:narrow(1,1,math.min(ord:numel(),max_per_image))
      keep = keep:index(1,ord)
      aboxes[j][i] = scored_boxes:index(1,keep)
    else
      aboxes[j][i] = torch.FloatTensor()
    end

    -- remove low scoring boxes and update threshold
    if i%1000 == 0 then
      aboxes[j],thresh[j] = keep_top_k(aboxes[j],max_per_set)
    end

  end
end


function pose_utils.finalPrunning(self, aboxes, thresh)
  local dataset = self.dataset
  for i = 1,dataset.num_classes do
    -- go back through and prune out detections below the found threshold
    for j = 1,dataset:size() do
      if aboxes[i][j]:numel() > 0 then
        local I = aboxes[i][j]:select(2,6):lt(thresh[i])
        local idx = torch.range(1,aboxes[i][j]:size(1)):long()
        idx = idx[I]
        if idx:numel()>0 then
          aboxes[i][j] = aboxes[i][j]:index(1,idx)
        end
      end
    end
  end
end

function pose_utils.evaluate(self, aboxes)
  local dataset = self.dataset
  local ap = {}
  local avp = {}
  for i=1,dataset.num_classes do
    local cls = dataset.classes[i]
    ap[i], avp[i] = pose_utils.VOCevaldet_pose(dataset,aboxes[i],cls, self.nPoses, self.discretize)
  end
  ap = torch.Tensor(ap)
  avp = torch.Tensor(avp)
  return {AP=ap, AVP=avp}
end


function pose_utils.pruneDetectionsSeparate(self, aboxes, thresh, output, boxes, i, max_per_image, max_per_set)
  local nPoses = self.nPoses
  local discretization_step = self.discretization_step or nPoses
  local frcnn_detec = self.frcnn_detec
  local add_bg = self.add_bg
  self._scored_boxes = self._scored_boxes or torch.FloatTensor()
  local scored_boxes = self._scored_boxes
  local dataset = self.dataset
  local I = dataset:getImage(i)
  local classif_scores = frcnn_detec:detect(I, boxes)
  classif_scores = classif_scores:index(2,torch.LongTensor{1,2,3,5,6,7,8,10,12,15,19,20,21})
  -- do a NMS for each class, based on the scores from the classifier
  for j=1,dataset.num_classes do
    --local scores = output:narrow(2,(j-1)*nPoses+add_bg + 1, nPoses):sum(2):select(2,1)
    local scores = classif_scores:select(2,j+1)--add_bg)
    local _,best_angles = output:narrow(2,(j-1)*discretization_step+add_bg + 1, discretization_step):max(2)
    if discretization_step == 360 then
      best_angles:add(-1) -- 0 -> 359
    end
    -- only select detections with a score greater than thresh
    -- this avoid doing NMS on too many bboxes with low score
    local idx = torch.range(1,scores:numel()):long()
    local idx2 = scores:gt(thresh[j])
    idx = idx[idx2]
    scored_boxes:resize(idx:numel(),6)
    if scored_boxes:numel() > 0 then
      scored_boxes:narrow(2,1,4):index(boxes,1,idx)
      scored_boxes:select(2,6):copy(scores[idx2])
      scored_boxes:select(2,5):copy(best_angles[idx2])
    end
    local keep = nms(scored_boxes,0.3)
    if keep:numel()>0 then
      local _,ord = torch.sort(scored_boxes:select(2,6):index(1,keep),true)
      ord = ord:narrow(1,1,math.min(ord:numel(),max_per_image))
      keep = keep:index(1,ord)
      aboxes[j][i] = scored_boxes:index(1,keep)
    else
      aboxes[j][i] = torch.FloatTensor()
    end

    -- remove low scoring boxes and update threshold
    if i%1000 == 0 then
      aboxes[j],thresh[j] = keep_top_k(aboxes[j],max_per_set)
    end

  end
end

function pose_utils.pruneDetectionsRegression(self, aboxes, thresh, output, boxes, i, max_per_image, max_per_set)
  local nPoses = self.nPoses
  local manif_size = self.manif_size
  --local a_i = torch.linspace(0,90,manif_size):mul(math.pi/180):data()
  local a_i = torch.linspace(-60,60,manif_size):mul(math.pi/180):data()
  local add_bg = 1
  self._scored_boxes = self._scored_boxes or torch.FloatTensor()
  local scored_boxes = self._scored_boxes
  local dataset = self.dataset
  local ang_pred = output
  local best_angles = torch.FloatTensor(boxes:size(1))
  local best_angles_c = best_angles:data()

  local frcnn_detec = self.frcnn_detec
  local dataset = self.dataset
  local I = dataset:getImage(i)
  local classif_scores = frcnn_detec:detect(I, boxes)
  classif_scores = classif_scores:index(2,torch.LongTensor{1,2,3,5,6,7,8,10,12,15,19,20,21})

  -- do a NMS for each class, based on the scores from the classifier
  for j=1,dataset.num_classes do
    --local scores = oscore:select(2,j+add_bg)
    local scores = classif_scores:select(2,j+1)--add_bg)
    local angs = ang_pred[j]--output:narrow(2,(j-1)*nPoses+add_bg + 1, nPoses):max(2)
    local angs_c = angs:data()
    if manif_size == 2 then
      for cc = 0, angs:size(1)-1 do
        best_angles_c[cc] = (math.atan2(angs_c[2*cc+1],angs_c[2*cc])*180/math.pi)%360
      end
    else
      for cc = 0, angs:size(1)-1 do
        local y1 = 0
        local x1 = 0
        for dim = 0, manif_size-1 do
          y1 = y1 + angs_c[manif_size*cc+dim]*math.cos(a_i[dim])
          x1 = x1 + angs_c[manif_size*cc+dim]*math.sin(a_i[dim])
        end
        -- check the order
        best_angles_c[cc] = (math.atan2(x1,y1)*180/math.pi)%360
        --best_angles_c[cc] = (math.atan2(angs_c[2*cc+1],angs_c[2*cc])*180/math.pi)%360
      end
    end
    -- only select detections with a score greater than thresh
    -- this avoid doing NMS on too many bboxes with low score
    local idx = torch.range(1,scores:numel()):long()
    local idx2 = scores:gt(thresh[j])
    idx = idx[idx2]
    scored_boxes:resize(idx:numel(),6)
    if scored_boxes:numel() > 0 then
      scored_boxes:narrow(2,1,4):index(boxes,1,idx)
      scored_boxes:select(2,6):copy(scores[idx2])
      scored_boxes:select(2,5):copy(best_angles[idx2])
    end
    local keep = nms(scored_boxes,0.3)
    if keep:numel()>0 then
      local _,ord = torch.sort(scored_boxes:select(2,6):index(1,keep),true)
      ord = ord:narrow(1,1,math.min(ord:numel(),max_per_image))
      keep = keep:index(1,ord)
      aboxes[j][i] = scored_boxes:index(1,keep)
    else
      aboxes[j][i] = torch.FloatTensor()
    end

    -- remove low scoring boxes and update threshold
    if i%1000 == 0 then
      aboxes[j],thresh[j] = keep_top_k(aboxes[j],max_per_set)
    end

  end
end


return pose_utils
