tcolors = {'red','blue','green','yellow','magenta','black','orange','pink','cyan','red','blue','green'}
function visualize_detections(im,aboxes,cl_names, thresh)
  local ok = pcall(require,'qt')
  if not ok then
    error('You need to run visualize_detections using qlua')
  end
  require 'qttorch'
  require 'qtwidget'

  local x,y = im:size(3),im:size(2)
  local w = qtwidget.newwindow(x,y,"Detections")
  local qtimg = qt.QImage.fromTensor(im)
  w:image(0,0,x,y,qtimg)
  local fontsize = 15
  thresh = thresh or 0.3

  for cl_idx, boxes in ipairs(aboxes) do
  
    local boxes_thresh = aboxes[cl_idx]
    if boxes_thresh:numel() > 0 then
      local max_score = boxes_thresh:select(2,6)
      local ang = boxes_thresh:select(2,5)
      local num_boxes = boxes_thresh:size(1)
      local widths  = boxes_thresh[{{},3}] - boxes_thresh[{{},1}]
      local heights = boxes_thresh[{{},4}] - boxes_thresh[{{},2}]

      for i=1,num_boxes do
        if max_score[i] > thresh then
          local x,y = boxes_thresh[{i,1}],boxes_thresh[{i,2}]
          local width,height = widths[i], heights[i]
          
          -- add bbox
          w:rectangle(x,y,width,height)
          
          -- add score
          w:moveto(x,y+fontsize)
          --w:setcolor("red")
          w:setcolor(tcolors[cl_idx])
          w:setfont(qt.QFont{serif=true,italic=true,size=fontsize,bold=true})
          if cl_names then
            w:show(string.format('%s: %.2f %.1f',cl_names[cl_idx],max_score[i], ang[i]))
          else
            w:show(string.format('%d: %.2f %.1f',cl_idx,max_score[i], ang[i]))
          end
        end
      end
      w:setcolor(tcolors[cl_idx])
      w:setlinewidth(2)
      w:stroke()
    end
  end
  return w
end

require 'nnf';

cls = {'aeroplane','bicycle','boat','bottle','bus','car',
              'chair','diningtable','motorbike',
              'sofa','train','tvmonitor'}

dsv = nnf.DataSetPascal{
  image_set='val',
  datadir='datasets/VOCdevkit',
  dataset_name='Pascal3D',
  --roidbdir='data/selective_search_data',
  roidbfile='data/MAD-DM2k/voc_2012_val.t7',
  year=2012,
  classes=cls
}

--aboxes = torch.load('cachedir/frcnn-pose_onlyflip,ang_res=24,model=alexnet,ptype=discrete/test_iter40000/boxes_Pascal3D')
aboxes = torch.load('/home/francisco/work/projects/pose-estimation/cachedir/frcnn-pose,ang_res=24,hard_gt=f,model=alexnet,ptype=mixed,tflip=t/test_iter40000/boxes_Pascal3D')

inv_aboxes = {}
for i = 1, dsv:size() do
  local bboxes = {}
  for cl = 1, dsv.num_classes do
    bboxes[cl] = aboxes[cl][i]
  end
  inv_aboxes[i] = bboxes
end

--dofile 'experiments/visualize_dets.lua'

function viewit(i)
  visualize_detections(dsv:getImage(i),inv_aboxes[i], dsv.classes)
end

--i = 1
--visualize_detections(dsv:getImage(i),inv_aboxes[i], dsv.classes)



